// Copyright (c) 2009 Dionysus Blazakis <dion@semantiscope.com>. All rights reserved.
//
// See LICENSE.txt for details.

#include "pin.h"
#include <map>
#include <set>
#include <string>
#include <iostream>
#include <fstream>

KNOB<std::string> KnobModuleWhitelist(KNOB_MODE_APPEND, "pintool",
  "mw", "", "whitelist a module for block coverage tracking");

KNOB<std::string> KnobId(KNOB_MODE_WRITEONCE, "pintool",
  "id", "none", "a name to identify runs from different inputs");

KNOB<UINT32> KnobRun(KNOB_MODE_WRITEONCE, "pintool",
  "run", "0", "an integer id to differentiate multiple runs easily");


typedef std::set<ADDRINT> Blocks;

struct LoadedImage {
  std::string name_;
  ADDRINT low_;
  ADDRINT high_;
  Blocks *hits_;

  LoadedImage(const std::string &n = "", ADDRINT low = 0, ADDRINT high = 0, Blocks *h = NULL) : 
    name_(n), low_(low), high_(high), hits_(h) {};

  bool operator<(const LoadedImage &rhs) const {
    return low_ < rhs.low_;
  };

  void RecordHits(void) const {
    std::string fname = KnobId.ValueString() + "-" + 
                        KnobRun.ValueString() + "-" + 
                        name_ + ".blocks";
    FILE *fp = fopen(fname.c_str(), "wb");
    if (fp == NULL) {
      LOG("[-] Failed opening " + fname + " to record block coverage\n");
      return;
    }

    unsigned int endian = 0xdeadbeef;
    if (fwrite(&endian, sizeof(endian), 1, fp) != 1 ||
        fwrite(&low_, sizeof(low_), 1, fp) != 1 ||
        fwrite(&high_, sizeof(high_), 1, fp) != 1) {
      LOG("[-] fwrite failure in block coverage dump header\n");
      fclose(fp);
      return;
    }

    for (Blocks::iterator i = hits_->begin(); i != hits_->end(); ++i) {
      ADDRINT a = *i;
      if (fwrite(&a, sizeof(a), 1, fp) != 1) {
        LOG("[-] Failed writing block address in block coverage dump\n");
        fclose(fp);
        return;
      }
    }

    fclose(fp);
  }
};

typedef std::set<LoadedImage> LoadedImages;
LoadedImages images;

typedef std::set<std::string> ImageWhitelist;
ImageWhitelist whitelist;


VOID PIN_FAST_ANALYSIS_CALL Ana_BlockHit(Blocks *b, ADDRINT a) {
  b->insert(a);
}

VOID ImageLoad(IMG img, VOID *v) {
  std::string path = IMG_Name(img);
  ADDRINT low = IMG_LowAddress(img);
  ADDRINT high = IMG_HighAddress(img);
  std::string name;

  // This is a pretty weak way to manipulate a path but... 
  // meh, this isn't a masterpiece
  std::string::size_type last_slash = path.rfind('\\');
  if (last_slash == std::string::npos) {
    LOG("[-] \"" + path + "\" is an image name without a slash; I suck as a program\n");
    name = path;
  } else {
    name = path.substr(last_slash + 1);
  }

  // Only track the whitelisted modules
  if (whitelist.find(name) != whitelist.end()) {
    images.insert(LoadedImage(name, low, high, new Blocks()));
    LOG("[+] Whitelist image loaded: " + name + "\n");
  }
}

VOID ImageUnload(IMG img, VOID *v) {
  ADDRINT low = IMG_LowAddress(img);

  LoadedImages::iterator i = images.find(LoadedImage("", low));
  if (i != images.end()) {
    LoadedImage li = *i;
    LOG("[+] Whitelist image unloaded: " + li.name_ + "\n");
    li.RecordHits();
    delete li.hits_;
    images.erase(i);
  }
}


VOID Trace(TRACE trace, VOID *v) {
  for (BBL bbl = TRACE_BblHead(trace); BBL_Valid(bbl); bbl = BBL_Next(bbl)) {
    ADDRINT addr = BBL_Address(bbl);
    LoadedImages::iterator i = images.upper_bound(LoadedImage("", addr));
    i--;
    if (i != images.end() && i->low_ <= addr && addr <= i->high_) {
      BBL_InsertCall(bbl, IPOINT_ANYWHERE, (AFUNPTR)Ana_BlockHit, 
        IARG_FAST_ANALYSIS_CALL, 
        IARG_PTR, i->hits_, 
        IARG_ADDRINT, addr, 
        IARG_END);
    }
  }
}

VOID Fini(INT32 code, VOID *v) {
  for (LoadedImages::iterator i = images.begin(); i != images.end(); ++i) {
    LOG("[/] *Late* whitelist image unload: " + i->name_ + "\n");
    i->RecordHits();
    delete i->hits_;
  }
  images.clear();
  LOG("[+] Done!\n");
}

int main(int argc, char *argv[]) {
  if (PIN_Init(argc,argv)) {
    LOG("[-] PIN_Init failed!\n");
    return -1;
  }

  LOG("[+] Whitelisted images:\n");
  for (UINT32 i = 0; i < KnobModuleWhitelist.NumberOfValues(); ++i) {
    LOG("[+]   " + KnobModuleWhitelist.Value(i) + "\n");
    whitelist.insert(KnobModuleWhitelist.Value(i));
  }

  IMG_AddInstrumentFunction(ImageLoad, NULL);
  IMG_AddUnloadFunction(ImageUnload, NULL);
  TRACE_AddInstrumentFunction(Trace, NULL);
  PIN_AddFiniFunction(Fini, NULL);

  // Never returns
  PIN_StartProgram();

  return 0;
}
