// Copyright (c) 2009 Dionysus Blazakis <dion@semantiscope.com>. All rights reserved.
//
// See LICENSE.txt for details.

#include <windows.h>
#include <set>

#include <ida.hpp>
#include <idp.hpp>
#pragma warning(push)
#pragma warning(disable: 4018) // '<' : signed/unsigned mismatch
#pragma warning(disable: 4800) // 'int' : forcing value to bool 'true' or 'false'
#include <graph.hpp>
#pragma warning(pop)
#include <loader.hpp>
#include <kernwin.hpp>

typedef std::set<ea_t> EaSet;
typedef std::vector<ea_t> EaVector;

static EaSet hit_blocks;
static EaSet hit_funcs;
static EaVector hit_funcsv;
static ea_t image_base = 0x10000000;

#ifdef IDACOV_RECOLOR_NAVBAR
static nav_colorizer_t *old_colorizer = NULL;
#endif

//--------------------------------------------------------------------------
static uint32 idaapi hit_funcsv_sizer(void *obj) {
  EaVector *hfs = (EaVector *) obj;
  return hfs->size();
}

static char *idaapi hit_funcsv_getl(void *obj, uint32 n, char *buf) {
  EaVector *hfs = (EaVector *) obj;

  if (n == 0) {
    qsnprintf(buf, 64, "Function Address");
    return buf;
  }
  n--;
  ea_t ea = (*hfs)[n];
  qsnprintf(buf, 64, "%08x", ea);
  return buf;
}

static void idaapi hit_funcsv_enter(void *obj, uint32 n) {
  EaVector *hfs = (EaVector *) obj;
  ea_t ea = (*hfs)[n - 1];
  jumpto(ea);
}

static int color_ea(ea_t ea, bgcolor_t *color)
{
  *color = 0xffffff;

  EaSet::iterator i = hit_blocks.find(ea);
  if (i != hit_blocks.end()) {
    *color = 0xffa0a0;
  } else {
    i = hit_funcs.upper_bound(ea);
    i--;
    if (i != hit_funcs.end()) {
      func_t *f = get_func(*i);
      if (f != NULL && f->contains(ea)) {
        *color = 0xffe7e7;
      }
    }
  }

  return 2;
}

//--------------------------------------------------------------------------
static int idaapi idp_cb(void *, int code, va_list va)
{
  int rv = 0;

  switch (code)
  {
    case processor_t::get_bg_color:
      // Get item background color
      // in: ea_t ea, bgcolor_t *color
      // Returns: 1-not implemented, 2-color set
      // Plugins can hook this callback to color disassembly lines
      // dynamically
      {
        ea_t ea = va_arg(va, ea_t);
        bgcolor_t *c = va_arg(va, bgcolor_t *);
        rv = color_ea(ea, c);
      }
      break;
  }
  return rv;
}

//--------------------------------------------------------------------------
#ifdef IDACOV_RECOLOR_NAVBAR
uint32 idaapi hit_colorizer(ea_t ea, asize_t nbytes)
{
  area_t nb_area(ea, ea + nbytes); // TODO: - 1? Inclusive vs. exclusive
  EaSet::iterator i = hit_blocks.upper_bound(ea + nbytes);

  i--;
  if (i != hit_blocks.end() && nb_area.contains(*i)) {
    return 0xff0000;
  }
  return 0x000000;
}
#endif

//--------------------------------------------------------------------------
static void mark_rest_of_block(ea_t startEA)
{
  ea_t first = image_base + startEA;
  ea_t curr = first;
  bool done = false;

  hit_funcs.insert(get_func(first)->startEA);

  while (curr != BADADDR) {
    // If we are a target of a jump and are not the first instruction
    // in a block, we are not part of the current block, stop.
    if (curr != first) {
      xrefblk_t xb;
      for (bool ok = xb.first_to(curr, XREF_ALL); ok && xb.iscode; ok = xb.next_to()) {
        if (xb.type == fl_JF || xb.type == fl_JN) {
          done = true;
          break;
        }
      }
    }

    if (done) break;
    hit_blocks.insert(curr);

    ea_t next = BADADDR;
    xrefblk_t xb;
    for (bool ok = xb.first_from(curr, XREF_ALL); ok && xb.iscode; ok = xb.next_from()) {
      if (xb.type == fl_JF || xb.type  == fl_JN) {
        done = true;
        break;
      } else if (xb.type == fl_F) {
        next = xb.to;
      }
    }

    if (done) break;
    curr = next;
  }
}

//--------------------------------------------------------------------------
int idaapi init(void)
{
  // unload us if text mode
  if (callui(ui_get_hwnd).vptr == NULL)
    return PLUGIN_SKIP;
  return PLUGIN_OK;
}

//--------------------------------------------------------------------------
void idaapi term(void)
{
  unhook_from_notification_point(HT_IDP, idp_cb, NULL);
}

//--------------------------------------------------------------------------
void idaapi run(int /*arg*/)
{
  image_base = get_imagebase();

  char *fname = askfile_c(0, "results.blocks", "Coverage results file");
  if (fname == NULL) {
    msg("fname was NULL\n");
  } else {
    FILE *fp = qfopen(fname, "r");
    if (fp == NULL) {
      msg("Failed opening file\n");
    } else {
      ea_t addr;
      while(qfscanf(fp, "%08x", &addr) == 1) {
        mark_rest_of_block(addr);
      }
    }
    qfclose(fp);
  }

#ifdef IDACOV_RECOLOR_NAVBAR
  old_colorizer = (nav_colorizer_t *) callui(ui_set_nav_colorizer, hit_colorizer).vptr;

  // A hope and a prayer
  refresh_navband(true);
#endif

  // Color the disassembly
  hook_to_notification_point(HT_IDP, idp_cb, NULL);
  refresh_idaview_anyway();

  for(EaSet::iterator i = hit_funcs.begin(); i != hit_funcs.end(); ++i) {
    hit_funcsv.push_back(*i);
  }

  choose(0, 
    -1, -1, -1, -1,
    &hit_funcsv,
    64,
    hit_funcsv_sizer,
    hit_funcsv_getl,
    "IDACov: Covered Functions",
    -1,
    0,
    NULL,
    NULL,
    NULL,
    NULL,
    hit_funcsv_enter,
    NULL,
    NULL,
    NULL);
}

//--------------------------------------------------------------------------

char comment[] = "This plugin paints the covered blocks blue and covered functions light blue";

char help[] =
        "IDACov -- displaying coverage results";

//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize
  term,                 // terminate. this pointer may be NULL.
  run,                  // invoke plugin
  comment,
  help,
  "IDACov",
  "Alt+2"
};
